# -*-python-*-
# $Id: s60-locatr.py,v 1.20 2006/09/29 18:04:03 asc Exp $

__package__    = "s60-locatr"
__version__    = "0.2"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/locatr/"
__cvsversion__ = "$Revision: 1.20 $"
__date__       = "$Date: 2006/09/29 18:04:03 $"
__copyright__  = "Copyright (c) 2006 Aaron Straup Cope. Perl Artistic License."

#
# WARNING : SPAGHETTI CODE AHEAD. WELL, MAYBE NOT
# SPAGHETTI BUT NOODLEY. NO REFUNDS ACCEPTED (YET).
#

# TO DO :
#
# * move code into function-related blocks, possibly classes
# * line wrapping 
# * automagic slideshow mode
# * do a sanity check on the error handling/reporting - it's
#   gotten a little out of control
# * redraw screen (title) correctly when requesting next photo
#   from the options menu
# * throw proper exceptions in zonetag.py to better see what
#   is going on

import e32
import appuifw
import socket
from key_codes import *

from elementtree.ElementTree import XML
from pdis.xpath import compile

if not e32.in_emulator() :
    import sysinfo
    import graphics
    import location
    
import zonetag
import anydbm
import os.path
import httplib
import urllib
import md5
import random
import time
import math

if e32.in_emulator() :
    import pwd
    import posix

#
#
#

class keyboard(object):
    
    def __init__(self,onevent=lambda:None):
        self._keyboard_state={}
        self._downs={}
        self._onevent=onevent

    def handle_event(self,event):
        if event['type'] == appuifw.EEventKeyDown:
            code=event['scancode']
            if not self.is_down(code):
                self._downs[code]=self._downs.get(code,0)+1
            self._keyboard_state[code]=1
        elif event['type'] == appuifw.EEventKeyUp:
            self._keyboard_state[event['scancode']]=0
        self._onevent()

    def is_down(self,scancode):
        return self._keyboard_state.get(scancode,0)

    def pressed(self,scancode):
        if self._downs.get(scancode,0):
            self._downs[scancode]-=1
            return True
        return False

#
#
#

class locatr :

    def __init__ (self) :

        # Flickr
        
        self.__flickr = {'www':'www.flickr.com',
                         'api':'api.flickr.com',
                         'static':'static.flickr.com',
                         'sz':'t'}
        
        self.__apikey = u"787acbb6cee3864e303499ea7a36f5d0"

        # Configs

        self.__approot = os.path.abspath("c")

        if e32.in_emulator() :
            user = pwd.getpwuid(posix.geteuid())
            self.__approot = os.path.abspath(user[5])

        self.__cfg     = None
        self.__cfgpath = u"c:\\locatr.cfg"

        if e32.in_emulator() :
            self.__cfgpath = os.path.abspath(os.path.join(self.__approot, u"locatr.db"))
        
        # BT socket fu
        
        self.__pair = None
        self.__sock = None
        self.__file = None

        # Logging
        
        self.__msg       = u""
        self.__msg_fetch = u"Fetching photo from Flickr..."
        
        self.__msglog  = []
        self.__verbose = False
        
        # Slideshow

        self.__canvas   = None
        self.__img      = None
        self.__photo    = None
        self.__coords   = None
        self.__pause    = False
        self.__kill     = False
        self.__photos   = []
        self.__tmpfiles = []

        self.__latlon = None
        self.__pages  = {}
        
        self.__fetching = False
        self.__source = ''

        self.__nogps = False
        
        # App

        self.lock  = e32.Ao_lock()        
        appuifw.app.exit_key_handler = self.abort
        
        self.__options = []
        
        #

        self.__kb = keyboard()

        self.__xstart = 10
        self.__ystart = 10            
        self.__ypos   = self.__ystart
        self.__yinc   = 15

        if not e32.in_emulator() :
            self.__canvas = appuifw.Canvas(event_callback=self.__kb.handle_event, redraw_callback=self.redraw_handler)
            self.__canvas.clear()

            appuifw.app.body = self.__canvas
            appuifw.app.screen = "large"

            (cx, cy) = self.__canvas.size
            self.__canvas_sz = (cx, cy)
                    
    #
    #
    #
    
    def loop (self) :

        if self.setup() :
            self.run()

        self.lock.wait()

    #
    #
    #
    
    def abort(self) :
        self.__kill = True
        
        self.disconnect_gps_device()
        self.store_cfg()
        self.cleanup()        
        self.lock.signal()

    #
    #
    #

    def cleanup (self, redraw=True) :
        self.__img    = None
        self.__photo  = None
        self.__coords = None
        self.__msglog = []
        self.__ypos   = self.__ystart

        if redraw :
            self.redraw_handler(None)
        
        if len(self.__tmpfiles) == 0 :
            return
        
        self.log("Cleaning up %s temporary files" % len(self.__tmpfiles))

        self.purge_cachefiles(self.__tmpfiles)
        self.log("All done")

    #
    #
    #
    
    def purge_cachefiles (self, files) :

        self.verbose("Clean up %s temp files" % len(files))
        
        while len(files) > 0 :
            upath = unicode(files.pop(0))
            self.verbose("Unlink %s" % upath)
            
            try :
                os.unlink(upath)
            except Exception , e :
                self.verbose("Failed to unlink %s, %s" % (upath, e))
                             
    #
    #
    #
    
    def setup (self) :

        if not self.open_cfg() :
            return False

        #

        zt = appuifw.query(u"Use ZoneTag for all queries", "query")
        
        if zt  :
            self.__nogps = True
        else :
            if not self.test_gps_connection() :
                self.log("All location queries will be done via ZoneTag.")
                self.__nogps = True

        #
        
        if not self.test_http_connection() :
            return False

        #
        
        self.assign_slideshow_menu()
        self.assign_gps_menu()
        self.assign_zonetag_menu()
        self.assign_verbose_menu((u"Enable verbose mode", self.toggle_verbose), True)            

        #

        (cx, cy) = self.canvas_size()

        if cx > 100 and cy > 100 :
            self.__flickr['sz'] = "m"

        if cx > 240 and cy > 240 :
            big = appuifw.query(u"You have a fancy display. Show big photos?", "query")
                
            if big != None :
                self.__flickr['sz'] = ''
            
        #

        return True

    #
    #
    #

    def test_gps_connection (self) :

        if e32.in_emulator() :
            return True

        self.log("Testing GPS connection");

        if not self.connect_gps_device() :
            self.log("GPS connection failed.")            
            return False

        return True
    
    #
    #
    #

    def test_http_connection (self) :

        self.log("Testing network connection");
        
        if not e32.in_emulator() :
            if sysinfo.signal() == 0 :
                self.error("No carrier signal")
                return False

        head = self.execute_http_request(self.__flickr['www'], '/', 'HEAD')
        
        if head == False :
            self.error("Unable to connect to flickr.com")
            return False
        
        return True
    
    #
    #
    #

    def open_cfg (self) :
        self.log("Reading configs")

        try :
            self.__cfg = anydbm.open(self.__cfgpath, "c")
        except Exception, e:
            self.error("Failed to open config file, %s" % e)
            return False
        
        return True

    #
    #
    #

    def store_cfg (self) :
        if self.__cfg != None :
            self.__cfg.close()
        
    #
    #
    #
    
    def run (self) :            
        self.log("Ready to show geotagged photos from Flickr!")
        self.log("To begin, select 'Show me photos!' from the Options menu.")
        
    #
    #
    #
    
    def slideshow (self, remove_photos_menu=False) :
        
        if remove_photos_menu :
            self.remove_photo_menus()
            
        self.cleanup(remove_photos_menu)
        
        res = self.poll()

        if res == None :
            self.error("Unable to determine current location.")
            return False

        lat = res[0]
        lon = res[1]
        
        photos = self.fetch_images(lat, lon)

        if len(photos) == 0 :
            self.error("No photos found for this spot : %s, %s" % (lat, lon))
            return False

        self.__latlon = res
        self.__photos = photos

        self.log("Showing %s Flickr photos near %s" % (len(photos), self.prettyprint_latlon(lat, lon)))
        
        self.show_one_photo(True)        
        return True

    #
    #
    #

    def open_photo_in_browser (self) :
        url = self.__photo['link']
        mob = appuifw.query(u"Open in Flickr mobile site", "query")

        if mob != None :
            url = self.__photo['moblink']
            
        self.open_url(url)

    #
    #
    #
    
    def open_url (self, url) :
        if e32.in_emulator() :
            self.log("Open URL %s" % url)
            return
        
        apprun  = 'z:\\system\\programs\\apprun.exe'
        browser = 'z:\\System\\Apps\\Browser\\Browser.app'
        e32.start_exe(apprun, browser + ' "%s"' % url , 1)

    #
    #
    #
    
    def show_one_photo (self, assign_menu=False) :

        if len(self.__photos) == 0 :

            if not appuifw.query(u"Collect more photos?", "query") :
                self.remove_photo_menus()
                return False

            return self.slideshow(True)

        #
        
        if assign_menu == True :
            self.assign_photo_menus()

        #

        if self.__fetching :
            self.error("Slow down! I am already getting the next photo.")
            return False

        #
        
        self.__fetching = True
        
        if self.__img != None :
            self.__photo['title'] = self.__msg_fetch
            self.redraw_handler(None, True)
        else :
            self.log(self.__msg_fetch)
            
        #

        try :
            ph  = self.__photos.pop()
            url = ph['url']
        except Exception, e :
            self.error("OMGWTF, %s" % e)
            self.__fetching = False            
            return False
        
        data = self.retrieve_image(url)

        if data == None :
            self.__fetching = False
            return False

        store = self.storage_path(url)

        if not self.store_image(store, data) :
            self.__fetching = False            
            return False

        img = self.load_image(store)

        if img == None :
            self.error("Failed to load %s" % store)
            self.__fetching = False            
            return False
        
        if e32.in_emulator() :
            self.log("render %s" % store)
        else :
            self.render_image(img)

        self.__fetching = False

        #

        center = self.prettyprint_latlon(self.__latlon[0], self.__latlon[1])
        
        dist   = self.calculate_distance(self.__latlon[0], self.__latlon[1], ph['lat'], ph['lon'])    
        bear   = self.calculate_bearing(self.__latlon[0], self.__latlon[1], ph['lat'], ph['lon'])

        ph['context'] = "%s miles %s from %s" % (dist, bear, center)
        ph['date']    = "Taken on %s" % ph['dt'].split(" ")[0]
        
        #
        
        self.set_title(ph['title'], ph['date'], ph['context'])
        
        #
        
        uid = self.digesterize(ph['id'])
        
        self.__cfg[uid] = unicode(time.time())
        self.__photo    = ph

        self.watch_keyboard_events()            
        return True

    #
    #
    #
    
    def set_title (self, title, dt='', ctx='') :

        if self.__fetching and title != self.__msg_fetch :
            return
        
        if e32.in_emulator() :
            for str in (title, dt, ctx) :
                self.log("%s" % str)
            return

        (cx, cy) = self.canvas_size(45)
        
        self.__canvas.text((self.__xstart, cy), unicode("\"%s\"" % title))
        self.__canvas.text((self.__xstart, cy + 15), unicode(dt))
        self.__canvas.text((self.__xstart, cy + 30), unicode(ctx))                
        
    #
    #
    #
    
    def watch_keyboard_events (self) :

        if self.__kb.pressed(EScancodeUpArrow) :
            return self.open_photo_in_browser()

        if self.__kb.pressed(EScancodeLeftArrow) :
            return self.slideshow(True)

        if self.__kb.pressed(EScancodeRightArrow) :
            return self.show_one_photo(False)

        e32.ao_sleep(1, self.watch_keyboard_events)
        
    #
    #
    #

    def connect_gps_device (self) :

        if not self.find_gps_device() :
            return False

        if not self.connect_gps_device_bt() :
            return False

        return True

    #
    #
    #
    
    def disconnect_gps_device (self) :

        if self.__sock == None :
            return True

        try :
            self.__sock.close()
        except Exception, e :
            self.error("Failed to disconnect from GPS device, %s" % e)
            return False

        self.__file = None
        self.__sock = None
        return True
    
    #
    #
    #

    def fetch_images (self, lat, lon, acc=16) :
        photos = []

        sz = 2
        
        if self.__source == 'zt' :
            sz = 4
            
        boundingbox = self.mk_bounding_box(lat, lon, sz)

        if boundingbox == None :
            return photos
        
        # self.log("bbox : %s (%s) " % (boundingbox, sz))

        cur_page  = 1
        num_pages = None

        sort_options = ('date-posted-asc',
                        'date-posted-desc',
                        'date-taken-asc',
                        'date-taken-desc',
                        'interestingness-desc',
                        'interestingness-asc',
                        'relevance')

        r = random.randint(1, len(sort_options)) - 1
        sort_by = sort_options[r]

        #
        
        page_key = self.digesterize("%s%s%s" % (boundingbox, acc, sort_by))

        if self.__pages.has_key(page_key) :
            cur_page = self.__pages[page_key]

        #
        
        while len(photos) < 10 :

            rem = 10 - len(photos)
            msg = ""

            if cur_page == 1 :
                self.log("Retrieving 10 photos, sorted by %s, from Flickr." % sort_by)
            else :
                self.log("Some of those have been viewed already.")
                self.log("Retrieving another %s photos, sorted by %s, from Flickr." % (rem, sort_by))
                       
            xml = self.api_call('flickr.photos.search',
                                bbox=boundingbox,
                                accuracy=acc,
                                page=cur_page,
                                per_page=10,
                                sort=sort_by,
                                min_upload_date=892800,
                                extras='geo,date_taken')

            #
            
            if num_pages == None :
                xp = compile("/rsp/photos")
                res = xp.evaluate(xml)
                num_pages = res[0].attrib['pages']
                
            if int(num_pages) == 0 :
                self.log("no pages")
                return []
            
            #
            
            xp  = compile("/rsp/photos/photo")
            res = xp.evaluate(xml)
        
            for ph in res :

                url = "/%s/%s_%s_%s.jpg" % (ph.attrib['server'] , ph.attrib['id'], ph.attrib['secret'], self.__flickr['sz'])
                
                if self.__flickr['sz'] == '' :
                    url = "/%s/%s_%s.jpg" % (ph.attrib['server'] , ph.attrib['id'], ph.attrib['secret'])                    

                id  = ph.attrib['id']
                uid = self.digesterize(id)
                
                if self.__cfg.has_key(uid) :
                    self.verbose("already seen %s" % url)
                    continue
                
                lat   = ph.attrib['latitude']
                lon   = ph.attrib['longitude']
                dt    = ph.attrib['datetaken']
                title = ph.attrib['title']

                link    = "http://%s/photos/%s/%s" % (self.__flickr['www'], ph.attrib['owner'], ph.attrib['id'])
                moblink = "http://%s/mob/photo.gne?id=%s" % (self.__flickr['www'], ph.attrib['id'])
                
                photos.append({'url':url,'link':link,'moblink':moblink, 'lat':lat,'lon':lon,'title':title,'id':id,'dt':dt})
                    
                if len(photos) == 10 :
                    break

            #
            
            cur_page += 1
            self.__pages[page_key] = cur_page

            #
            
            self.verbose("pages : %s / %s" % (cur_page, num_pages))
            
            if int(cur_page) > int(num_pages) :
                break

        #
        #
        #

        return photos
    
    #
    #
    #
    
    def poll (self) :

        self.log("Polling for location data.")
        
        self.__source = ''
        
        if e32.in_emulator() :
            self.__source = 'emulator'
            return (37.750917, -122.421823)

        res = None

        if not self.__nogps :
            res = self.poll_gps()

        if res == None :
            self.verbose("Polling ZoneTag for location data")
            res = self.poll_zonetag()

        self.verbose("Finished polling")
        return res
    
    #
    #
    #

    def poll_zonetag (self) :
        self.__source = 'zt'

        if self.configure_zonetag(False) :
            zt = zonetag.zonetag(self.__cfg['zt_apptoken'], self.__cfg['zt_usertoken'])
            self.verbose("Created ZoneTag object")
            
            (mcc, mnc, lac, cellid) = self.identify_celltower()
            self.verbose("Tower : %s %s %s %s" % (mcc, mnc, lac, cellid))
            
            res = None
            
            try :
                res = zt.geocode_celltower(mcc, mnc, lac, cellid)
            except Exception, e :
                self.error("Failed to perform ZoneTag magic:")
                self.error(e)
            return res
        
    #
    #
    #

    def configure_zonetag (self, force=True) :

        if not self.__cfg.has_key('zt_apptoken') or self.__cfg['zt_apptoken'] == '' or force :
            apptoken  = appuifw.query(u'Your ZoneTag application token', 'text')

            if apptoken == None :
                return False

            self.__cfg['zt_apptoken']  = apptoken

        #
        
        if not self.__cfg.has_key('zt_usertoken') or self.__cfg['zt_usertoken'] == '' or force :
            usertoken = appuifw.query(u'Your ZoneTag user token', 'text')

            if usertoken == None :
                return False
            
            self.__cfg['zt_usertoken'] = usertoken

        return True
    
    #
    #
    #
    
    def poll_gps (self) :
        self.__source = 'gps'
        
        if self.__sock == None :
            if not self.connect_gps_device() :
                self.log("Unable to connect to GPS device.")
                return None

        #

        data = self.readlines("$GPRMC")

        lat = None
        lon = None
        
        if data != "" :
            self.verbose(data)

            try :
                (lat, lon) = self.parse_gps_data(data)
                self.verbose("%s, %s" % (lat, lon))
                
            except Exception, e :
                self.error("GPS parse error, %s" % e)
                return None
            
        #
        
        self.disconnect_gps_device()
        return (lat, lon)

    #
    #
    #
    
    def mk_bounding_box (self, lat, lon, sz=2) :

        self.verbose("mk bbox : %s, %s, %s" % (lat, lon, sz))

        """
        http://jehiah.com/archive/spatial-proximity-searching-using-latlongs

        baz = acos(sin(radians(latitude))
                   * sin( radians(%s))
                   + cos(radians(latitude))
                   * cos( radians(%s))
                   * cos( radians(longitude - %s) )
        )
        
        bar = degrees(baz)
        foo = (bar * 69.09)
        
        truncate(foo,1)

        as distance
        
        SELECT user_id
        FROM users, zip_codes
        WHERE users.zip = zip_codes.zip
        AND degrees(acos(
            sin( radians(zip_codes.latitude) )
            * sin( radians(47.604718))
            + cos( radians(zip_codes.latitude))
            * cos( radians(47.604718))
            * cos( radians(zip_codes.longitude - -122.335230) )
            ) ) * 69.09 < 50
        """
        
        edge   = float(sz) / float(2)
        offset = float(edge) / float(69)

        self.verbose("bbox offset %s" % offset)
        
        try :
            lat = float(lat)
            lon = float(lon)
        except Exception, e :
            self.error("Failed to format lat, long values : %s" % e)
            return None
        
        swlat = float("%.6f" % (lat - offset))
        swlon = float("%.6f" % (lon - offset))
        nelat = float("%.6f" % (lat + offset))
        nelon = float("%.6f" % (lon + offset))

        # d = self.distance(swlat, swlon, nelat, nelon)
        # self.log("bbox dist : %s" % d)
        
        return "%s,%s,%s,%s" % (swlon, swlat, nelon, nelat)
        
    #
    #
    #
    
    def parse_gps_data (self, gps_data) :

        # Thank you :
        # http://positio.rista.net/en/pys60gps/
        
        # Expect 11 ','-delimiters (NMEA 2.2)
        # Expect 12 ','-delimiters (NMEA 2.3)
        
        if (gps_data.count(",") == 11) :
            (GPRMC,hhmmssss,status,lat,l1,lon,l2,speed_knots,course,ddmmyy,mvd,mdr_cs)=gps_data.split(",")
        elif (gps_data.count(",") == 12) :
            (GPRMC,hhmmssss,status,lat,l1,lon,l2,speed_knots,course,ddmmyy,mvd,mdr,fix_cs)=gps_data.split(",")
        else:
            self.error("Invalid GPRMC string")

        if (status is 'A' and lat is not "" and lon is not ""):
            lat, lon = self.parse_gps_latlong(lat, l1, lon, l2)

        return (lat, lon)

    #
    #
    #

    def parse_gps_latlong (self, lat, l1, lon, l2):
        
        try:
            latitude = float(lat[0:2]) + (float(lat[2:4] + "." + lat[5:9])/60)
            longitude = float(lon[0:3]) + (float(lon[3:5] + "." + lon[6:10])/60)

            if l1 is "S":
                latitude = -latitude

            if l2 is "W":
                longitude = -longitude

            latitude  = float("%.6f" % latitude)
            longitude = float("%.6f" % longitude)
            
            return latitude, longitude
        except:
            return (None, None)
        
    #
    #
    #
    
    def readlines (self, type, iterations=50) :
        self.verbose("Whispering sweet nothings...")
        
        count = 0
        data  = ""
        
        while count < iterations :

            try :
                ln = self.__file.readline()
                ln.strip()
                
                if ln.startswith(type) :
                    data = ln
                    break

            except socket.error, e :
                self.error("There was an error talk to the GPS device : %s" % e)
                break
            except Excpetion, e :
                self.error("Hrm - I have no idea what's going on : %s" % e)
                break
            
            count += 1

        if data == "" :
            self.error("Failed to receive any suitable data after %s iterations" % iterations)
            
        return data
    
    #
    #
    #
        
    def find_gps_device (self) :

        self.verbose("finding GPS device")
        
        if self.__cfg.has_key('gps_address') :
            self.verbose("Using default GPS device")
            
            self.__pair = (self.__cfg['gps_address'], int(self.__cfg['gps_service']))
            return True

        return self.pair_gps_device()
        
    #
    #
    #
    
    def pair_gps_device (self) :
        self.verbose("Pairing with unknown GPS device")
        
        try :
            address, services=socket.bt_discover()
        except Exception, e :
            self.error("Failed to discover GPS device, %s" % e)
            return False

        service = self.choose_bt_service(services)

        if service == None :
            self.error("Failed to discover GPS service")
            return False

        save = appuifw.query(u"Store as default GPS device?", "query")

        if save != None :
            self.__cfg['gps_address'] = unicode(address)
            self.__cfg['gps_service'] = unicode(service)

            self.log("store address %s" % address)
            self.log("store serivce %s" % service)

        self.__pair = (address, service)            

        self.verbose("Paired with %s" % address)
        return True

    #
    #
    #
    
    def choose_bt_service (self, services) :

        if len(services)>1:
            choices=services.keys()
            choices.sort()

            choice=appuifw.popup_menu([unicode(services[x])+": "+x for x in choices], u'Choose GPS Service:')

            if choice == None :
                return None
            
            return services[choices[choice]]
        
        else:
            return services.values()[0]
        
    #
    #
    #

    def connect_gps_device_bt (self) :

        self.verbose("Connecting to GPS device")
        
        try :
            sock=socket.socket(socket.AF_BT,socket.SOCK_STREAM)
        except Exception, e :
            self.error("Failed to create socket, %s" % e)
            return False

        try : 
            sock.connect(self.__pair)
        except Exception, e :
            self.error("Failed to connect to GPS device, %s" % e)
            return False

        self.__sock = sock
        self.__file = self.__sock.makefile(512)
        
        return True

    #
    #
    #

    def api_call(self, method, **kwargs) :

        kwargs['method']  = method
        kwargs['api_key'] = self.__apikey
        
        url = "/services/rest/?%s" % urllib.urlencode(kwargs)

        self.verbose(url)

        res = self.execute_http_request(self.__flickr['api'], url)

        if res == None :
            return None
        
        raw = res.read()
        # self.verbose(raw)

        try :
            xml = XML(raw)
        except Exception, e :
            self.log("XML parse error : %s" % e)
            return None

        if self.is_api_error(xml) :
            return None

        return xml

    #
    #
    #
    
    def is_api_error (self, xml) :

        xp  = compile("/rsp")
        rsp = xp.evaluate(xml)[0]

        if rsp.attrib['stat'] == 'fail' :
            err = rsp.find('err')
            self.log("API error : %s (%s)" % (err.attrib['code'], err.attrib['msg']))
            return True

        return False

    #
    #
    #
    
    def execute_http_request (self, endpoint, url, method='GET') :

        try :
            conn = httplib.HTTPConnection(endpoint)
            conn.request(method, url)
        except Exception, e :
            self.log("HTTP request failed : %s" % e)
            return None
            
        return conn.getresponse()
    
    #
    #
    #

    def toggle_verbose (self) :

        if self.__verbose :
            self.__verbose = False
            self.assign_verbose_menu((u"Enable verbose mode", self.toggle_verbose))            
        else :
            self.__verbose = True
            self.assign_verbose_menu((u"Disable verbose mode", self.toggle_verbose))

    #
    #
    #

    def assign_slideshow_menu (self, startup=False) :
        self.__options.insert(0, (u"Show me photos!", self.slideshow))
        self.assign_menu()

    #
    #
    #

    def assign_gps_menu (self) :
        if not e32.in_emulator() :
            self.__options.append((u'Pair with GPS device', self.pair_gps_device))
            self.assign_menu()

    #
    #
    #
    
    def assign_photo_menus (self) :

        # FIX ME : this shouldn't be necessary
        # but it keeps being invoked by keyboard
        # events and the default trigger (in the
        # method signature) is not honoured
        
        if self.__options[1][0] != "Next photo" :
            self.__options.insert(1, (u"Next photo", self.show_one_photo))
            self.__options.insert(2, (u"Open photo", self.open_photo_in_browser))        
            self.assign_menu()

    #
    #
    #

    def assign_zonetag_menu (self) :
        self.__options.append((u'Configure ZoneTag', self.configure_zonetag))
    #
    #
    #
    
    def remove_photo_menus (self) :
        self.__options = self.__options[3:]
        self.assign_slideshow_menu()
            
    #
    #
    #
    
    def assign_verbose_menu (self, menu_item, startup=False) :

        if startup :
            self.__options.append(menu_item)
        else :
            idx = len(self.__options) - 1
            self.__options[idx] = menu_item

        self.assign_menu()

    #
    #
    #

    def assign_menu (self) :
        appuifw.app.menu = self.__options
        
    #
    #
    #
    
    def verbose (self, msg) :

        if not self.__verbose :
            return

        self.log(msg)

    #
    #
    #

    def error (self, msg) :

        if e32.in_emulator() :
            self.log(msg)
            return

        self.log("[ERROR] %s" % msg)
        # appuifw.note(unicode(msg), "error")        
        
    #
    #
    #
    
    def log (self, msg) :

        if e32.in_emulator() :
            self.__msg += unicode("%s\n" % msg)
            self.write(self.__msg)
            return

        ypos = self.__ypos + self.__yinc
        (cx, cy) = self.canvas_size(0)

        if (ypos + 10) >= cy :
            self.canvas_clear()
            self.__msglog = []
            ypos = self.__ystart

        self.__ypos = ypos
        umsg = unicode(msg)
        
        self.__canvas.text((self.__xstart, ypos), umsg)
        self.__msglog.append((ypos, umsg))

    #
    #
    #
    
    def relog (self) :
        self.canvas_clear()
        
        for ypos, umsg in self.__msglog :
            self.__canvas.text((self.__xstart, ypos), umsg)
            
    #
    #
    #
    
    def write(self, txt) :
        txt = unicode(txt)
        
        if e32.in_emulator() :
            t = appuifw.Text()
            t.write(txt)
            appuifw.app.body = t

        else :
            appuifw.app.body = appuifw.Text(txt)

    #
    #
    #

    def storage_path (self, url) :
        hex   = self.digesterize(url)
        fname = u"fl-%s.jpg" % hex

        store = u"c:\\%s" % fname

        if e32.in_emulator() :
            store = os.path.abspath(os.path.join(self.__approot, fname))
            
        return store

    #
    #
    #

    def digesterize (self, url) :
        return md5.md5(url).hexdigest()
    
    #
    #
    #

    def retrieve_image (self, path) :
        self.verbose("retrieve %s" % path)
        
        data = self.execute_http_request(self.__flickr['static'], path)
        return data

    #
    #
    #

    def load_image (self, path) :

        if not self.carry_on() :
            return None

        self.verbose("load image, %s" % path)
        
        if e32.in_emulator() :
            return True
        
        try :
            image = self.open_image(path)
        except Exception, e :
            self.error("Failed to open image %s, %s" % (store, e))
            return None

        try :
            image = self.resize_image(image)
        except Exception, e :
            self.error("Failed to resize image, %s" % e)
            return None
            
        return image

    #
    #
    #

    def store_image (self, path, data) :

        if not self.carry_on() :
            return False

        self.verbose("store %s" % path)
        
        try :
            fh = open(path, "w")
            fh.write(data.read())
            fh.close()
        except Exception, e :
            self.error("Failed to store image locally in %s, %s" % (path, e))
            return False

        if path not in self.__tmpfiles :
            self.__tmpfiles.append(path)

        return True

    #
    #
    #
    
    def open_image (self, path) :

        if not self.carry_on() :
            return None

        self.verbose("open %s" % path)
        return graphics.Image.open(path)

    #
    #
    #
    
    def resize_image (self, image) :

        if not self.carry_on() :
            return None

        self.verbose("resize image")
        
        (ix, iy) = image.size
        (cx, cy) = self.canvas_size()
        
        if ix <= cx and iy <= cy :
            return image

        if ix > iy :
            rx = int(cx)
            ry = int(cy * (float(cx) / float(ix)))
        else :
            rx = int(cx * (float(cy) / float(iy)))
            ry = int(cy)

        return image.resize((rx, ry), callback=None, keepaspect=1)

    #
    #
    #

    def centering_coords (self, image) :

        (ix, iy) = image.size
        (cx, cy) = self.canvas_size()
        
        if ix == cx and iy == cy :
            return (0, 0)
        
        rx = int(float(cx - ix) / 2)
        ry = int(float(cy - iy) / 2)        

        return (rx, ry)
    
    #
    #
    #

    def render_image (self, image) :

        self.verbose("Render image")
        
        if not self.carry_on() :
            return False

        coords = self.centering_coords(image)
        
        self.canvas_clear()
        self.blit(image, coords)
        
    #
    #
    #
    
    def blit(self, image, coords) :
        self.__img    = image
        self.__coords = coords

        self.__canvas.blit(image, target=coords)
        
    #
    #
    #
        
    def redraw_handler (self, rect, clear=False) :

        if not self.carry_on() :
            return None

        if self.__img != None :

            if clear :
                self.canvas_clear()
                
            self.blit(self.__img, self.__coords)
            self.set_title(self.__photo['title'], self.__photo['date'], self.__photo['context'])
        else :
            self.relog()
            
    #
    #
    #

    def canvas_clear (self) :
        if e32.in_emulator() :
            return

        self.__canvas.clear()

    #
    #
    #
    
    def canvas_size (self, adjusted=50) :

        if e32.in_emulator() :
            return (240, 180)
        
        (cx, cy) = self.__canvas_sz

        if adjusted > 0 :
            cy -= adjusted

        return (cx, cy)

    #
    #
    #

    def identify_celltower (self) :

        if e32.in_emulator() :

            # Williamsburg
            return (310, 260, 32106, 29682)
        
            # SFO
            return (310, 170, 270, 21152)

        #
        
        return location.gsm_location()

    #
    #
    #

    def calculate_bearing (self, from_lat, from_lon, to_lat, to_lon) :

        from_lat = float(from_lat)
        from_lon = float(from_lon)
        to_lat   = float(to_lat)
        to_lon   = float(to_lon)

        NS = ""
        EW = ""

        if to_lat == from_lat :
            NS = ""
        elif to_lat < from_lat :
            NS = "S"
        else :
             NS = "N"

        if to_lon == from_lon :
            EW = ""
        elif to_lon < from_lon :
            EW = "W"
        else :
            EW = "E"
            
        return unicode(NS + EW)
    
        x = math.sin(to_lon - from_lon) * math.cos(to_lat)
        y = math.cos(from_lat) * math.sin(to_lat) - math.sin(from_lat) * math.cos(to_lat) * math.cos(to_lon - from_lon)
        return math.atan2(y, x)

    #
    #
    #
    
    def calculate_distance (self, from_lat, from_lon, to_lat, to_lon) :

        from_lat = self.rad(float(from_lat))
        from_lon = self.rad(float(from_lon))
        to_lat   = self.rad(float(to_lat))
        to_lon   = self.rad(float(to_lon))
        
        r = 3963.1
        dist = math.acos(math.sin(from_lat) * math.sin(to_lat) + math.cos(from_lat) * math.cos(to_lat) * math.cos(to_lon-from_lon)) * r
        return float("%.2f" % dist)

    #
    #
    #

    def prettyprint_latlon (self, lat, lon) :
        lat = float(lat)
        lon = float(lon)

        bear_lat = ""
        bear_lon = ""
        
        if lat < 0 :
            bear_lat = "S"
        elif lat > 0 :
            bear_lat = "N"
        else :
            pass
            
        if lon < 0 :
            bear_lon = "W"
        elif lon > 0 :
            bear_lon = "E"
        else :
            pass
            
        lat_dms = self.decimal_to_dms(lat)
        lon_dms = self.decimal_to_dms(lon)

        lat_str = "%d %d %.2f %s" % (lat_dms[0], lat_dms[1], lat_dms[2], bear_lat)
        lon_str = "%d %d %.2f %s" % (lon_dms[0], lon_dms[1], lon_dms[2], bear_lon)        

        return "%s %s" % (lat_str, lon_str)
       
    #
    #
    #
    
    def decimal_to_dms (self, dec) :

        deg = int(dec);

        dec_min = abs(dec - deg) * 60
        min     = int(dec_min)

        sec = (dec_min - min) * 60
        sec = float("%.2f" % sec)

        return (deg, min, sec)
        
    #
    #
    #
    
    def rad (self, v) :
        return v * math.pi / 180

    #
    #
    #
    
    def carry_on (self) :
        if self.__kill :
            return False
        
        return True

    #
    #
    #
    
if __name__ == "__main__" :
    app = locatr()
    app.loop()
